unit VShellTree;

// VShellTree 1.1.1
// Delphi 7 to 2010 (Win32 versions only).

//
//
// LICENSE
//
// The contents of this file are subject to the Mozilla Public License
// Version 1.1 (the "License"); you may not use this file except in
// compliance with the License. You may obtain a copy of the License
// at http://www.mozilla.org/MPL/.
//
// Alternatively, you may redistribute this library, use and/or modify
// it under the terms of the GNU Lesser General Public License as
// published by the Free Software Foundation; either version 2.1 of the
// License, or (at your option) any later version. You may obtain a
// copy of the LGPL at http://www.gnu.org/copyleft/.
//
// Software distributed under the License is distributed on an "AS IS"
// basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
// the License for the specific language governing rights and
// limitations under the License.
//
// The original code is DrawTreeDemo.pas, written by Mike Lischke.
//
// Most of the code has been modified or added by Dieter Khler
// (Heidelberg, Germany, www.philo.de). The derived code is
// VShellTree.pas.
//
// The initial developer of the original code is digital publishing AG
// (Munich, Germany, www.digitalpublishing.de), written by Mike Lischke
// (public@soft-gems.net, www.soft-gems.net).
//
// Portions created by digital publishing AG are Copyright
// (C) 1999-2001 digital publishing AG. All Rights Reserved.
//
// Portions created by Dieter Khler are Copyright (C) 2007-2008
// Dieter Khler. All Rights Reserved.

// HISTORY
//
// Only major modification are listed here.
//
// 2010-01-18 1.1.1 Made compatible with Virtual TreeView 4.8.6
// 2008-12-04 1.1.0 Made compatible with Virtual TreeView 4.7.0
// 2007-06-08 1.0.0

interface

uses
  Windows, ExtCtrls, SysUtils, Controls, Classes, Forms, ImgList, VirtualTrees;

type
  TVstvFileDblClickEvent = procedure(Sender: TObject; Path: WideString; Attributes: Cardinal) of object;

  TVstvSpecialFolder = (
    vsfDesktop, vsfInternet, vsfPrograms, vsfControls, vsfPrinters,
    vsfPersonal, vsfFavorites, vsfStartup, vsfRecent, vsfSendto,
    vsfBitBucket, vsfStartmenu, vsfDesktopDirectory, vsfDrives,
    vsfNetwork, vsfNethood, vsfFonts, vsfTemplates, vsfCommonStartMenu,
    vsfCommonPrograms, vsfCommonStartup, vsfCommonDesktopDirectory,
    vsfAppData, vsfPrintHood, vsfAltStartup, vsfCommonAltStartup,
    vsfCommonFavorites, vsfInternetCache, vsfCookies, vsfHistory );

  TVstvSpecialFolders = set of TVstvSpecialFolder;

  TVirtualShellTree = class(TCustomControl)
  private
    FSystemImages: TImageList;
    FVST: TVirtualStringTree;
    FOnFileDblClick: TVstvFileDblClickEvent;
    FTopFolders: TVstvSpecialFolders;
    procedure VSTInitNode(Sender: TBaseVirtualTree; ParentNode,
      Node: PVirtualNode; var InitialStates: TVirtualNodeInitStates);
    procedure VSTFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure VSTInitChildren(Sender: TBaseVirtualTree;
      Node: PVirtualNode; var ChildCount: Cardinal);
    procedure VSTGetImageIndex(Sender: TBaseVirtualTree;
      Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex;
      var Ghosted: Boolean; var ImageIndex: Integer);
    procedure VSTCompareNodes(Sender: TBaseVirtualTree; Node1,
      Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
    procedure VSTDblClick(Sender: TObject);
    procedure VSTGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType;
      var CellText: UnicodeString);
  protected
    procedure DoFileDblClick(const Path: WideString; Attributes: Cardinal); virtual;
    procedure Reset; virtual;
    procedure SetTopFolders(const Value: TVstvSpecialFolders); virtual;

    property SystemImages: TImageList read FSystemImages;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Align;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelEdges;
    property BevelKind;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property TopFolders: TVstvSpecialFolders read FTopFolders write SetTopFolders default [vsfPersonal, vsfDrives, vsfNetwork, vsfDesktopDirectory];
    property Visible;

    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnFileDblClick: TVstvFileDblClickEvent read FOnFileDblClick write FOnFileDblClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
  end;

procedure Register;

implementation

uses
dialogs,  ShellAPI, ShlObj, ComObj, ActiveX;

procedure Register;
begin
  RegisterComponents('Open XML',[TVirtualShellTree]);
end;

var
  Malloc: IMalloc;
  DesktopFolder: IShellFolder;

const
  CsidlMap: array[TVstvSpecialFolder] of LongWord = (
    CSIDL_DESKTOP, CSIDL_INTERNET, CSIDL_PROGRAMS, CSIDL_CONTROLS, CSIDL_PRINTERS,
    CSIDL_PERSONAL, CSIDL_FAVORITES, CSIDL_STARTUP, CSIDL_RECENT, CSIDL_SENDTO,
    CSIDL_BITBUCKET, CSIDL_STARTMENU, CSIDL_DESKTOPDIRECTORY, CSIDL_DRIVES,
    CSIDL_NETWORK, CSIDL_NETHOOD, CSIDL_FONTS, CSIDL_TEMPLATES,
    CSIDL_COMMON_STARTMENU, CSIDL_COMMON_PROGRAMS, CSIDL_COMMON_STARTUP,
    CSIDL_COMMON_DESKTOPDIRECTORY, CSIDL_APPDATA, CSIDL_PRINTHOOD,
    CSIDL_ALTSTARTUP, CSIDL_COMMON_ALTSTARTUP, CSIDL_COMMON_FAVORITES,
    CSIDL_INTERNET_CACHE, CSIDL_COOKIES, CSIDL_HISTORY);

type
  // This data record contains all necessary information about a particular file system object.
  // This can either be a folder (virtual or real) or a file.
  PShellObjectData = ^TShellObjectData;
  TShellObjectData = record
    ShellFolder: IShellFolder;  // The ShellFolder interface for the file.
    PIDL: PItemIDList;          // Pointer to the Item Identifier List of the file.
    Display: WideString;
    Attributes: Cardinal;
    OpenIndex,
    CloseIndex: Integer;        // Image indices into the system image list.
  end;

{ TVirtualShellTree }

constructor TVirtualShellTree.Create(AOwner: TComponent);
var
  SFI: TSHFileInfo;
begin
  inherited;
  FSystemImages := TImageList.Create(Self);
  FSystemImages.Handle := SHGetFileInfo('', 0, SFI, SizeOf(SFI), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  FSystemImages.ShareImages := True;

  FVST := TVirtualStringTree.Create(Self);
  with FVST do begin
    // Properties:
    Parent := Self;
    Align := alClient;
    BevelEdges := [];
    BevelInner := bvNone;
    BevelKind := bkNone;
    BevelOuter := bvNone;
    Images := FSystemImages;
    IncrementalSearch := isAll;
    Indent := 10;
    TreeOptions.AnimationOptions := [toAnimatedToggle];
    TreeOptions.AutoOptions := [toAutoDropExpand, toAutoScroll, toAutoScrollOnExpand, toAutoTristateTracking,
                                toAutoDeleteMovedNodes, toAutoFreeOnCollapse];
    TreeOptions.PaintOptions := [toHideSelection, toShowButtons, toShowDropmark, toShowRoot, toShowTreeLines,
                                 toThemeAware, toUseBlendedImages, toAlwaysHideSelection];
    // Events:
    OnCompareNodes := VSTCompareNodes;
    OnDblClick := VSTDblClick;
    OnFreeNode := VSTFreeNode;
    OnGetImageIndex := VSTGetImageIndex;
    OnGetText := VSTGetText;
    OnInitChildren := VSTInitChildren;
    OnInitNode := VSTInitNode;
  end;

  FTopFolders := [vsfPersonal, vsfDrives, vsfNetwork, vsfDesktopDirectory];

  FVST.NodeDataSize := SizeOf(TShellObjectData);
  Reset;
end;

procedure TVirtualShellTree.SetTopFolders(
  const Value: TVstvSpecialFolders);
begin
  if FTopFolders <> Value then begin
    FTopFolders := Value;
    Reset;
  end;
end;

procedure TVirtualShellTree.DoFileDblClick(const Path: WideString;
  Attributes: Cardinal);
begin
  if Assigned(FOnFileDblClick) then
    FOnFileDblClick(Self, Path, Attributes);
end;

procedure TVirtualShellTree.Reset;
var
  Count: Cardinal;
  I: Integer;
begin
  // Determine the number of items in the TopFolders set.
  Count := 0;
  for I := 0 to Pred(Length(CsidlMap)) do
    if (TVstvSpecialFolder(I) in TopFolders) then
      Inc(Count);

  if not (csDesigning in ComponentState) then begin
    FVST.Clear;
    FVST.RootNodeCount := Count;
  end;
end;

procedure TVirtualShellTree.VSTInitNode(Sender: TBaseVirtualTree;
  ParentNode, Node: PVirtualNode; var InitialStates: TVirtualNodeInitStates);
var
  Data: PShellObjectData;
  ItemListPointer: PItemIDList;
  ShellFolder: IShellFolder;
  ShellInfo : TSHFileInfo;
  Count: Cardinal;
  I: Integer;
begin
  Data := Sender.GetNodeData(Node);

  case Sender.GetNodeLevel(Node) of
    0:
    begin
      Count := 0;
      for I := 0 to Pred(Length(CsidlMap)) do
        if (TVstvSpecialFolder(I) in TopFolders) then begin
          Inc(Count);
          if Node.Index = Pred(Count) then begin
            // Retrieve the ItemIDList pointer for the special folder:
            OleCheck(SHGetSpecialFolderLocation(0, CsidlMap[TVstvSpecialFolder(I)], ItemListPointer));
            Data.PIDL := ItemListPointer;
            // Retrieve the ShellFolder interface for the special folder:
            if TVstvSpecialFolder(I) = vsfDesktop then begin
              Data.ShellFolder := DesktopFolder;
            end else
              if DesktopFolder.BindToObject(ItemListPointer, nil, IID_IShellFolder, ShellFolder) = NOERROR
                then Data.ShellFolder := ShellFolder
                else Data.ShellFolder := nil;

              // Get details about the node via SHGetFileInfo
              OleCheck(SHGetFileInfo(PChar(ItemListPointer), 0, ShellInfo,
                            SizeOf(ShellInfo), SHGFI_PIDL or SHGFI_DISPLAYNAME or SHGFI_ATTRIBUTES	or SHGFI_SYSICONINDEX or SHGFI_SMALLICON));
              Data.Display := ShellInfo.szDisplayName;
              Data.Attributes := ShellInfo.dwAttributes;
              Data.CloseIndex := ShellInfo.iIcon;
              SHGetFileInfo(PChar(ItemListPointer), 0, ShellInfo,
                            SizeOf(ShellInfo), SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_OPENICON);
              Data.OpenIndex := ShellInfo.iIcon;

            break;
          end;
        end;
    end;
  end;

  if (Data.Attributes and SFGAO_FOLDER) <> 0 then
    Include(InitialStates, ivsHasChildren);
end;

procedure TVirtualShellTree.VSTFreeNode(Sender: TBaseVirtualTree;
  Node: PVirtualNode);
var
  Data: PShellObjectData;
begin
  Data := Sender.GetNodeData(Node);
  if Assigned(Data) then begin
    if Assigned(Data.PIDL) then
      Malloc.Free(Data.PIDL);
    Finalize(Data^); // Clear string data.
  end;
end;

procedure TVirtualShellTree.VSTGetText(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
  var CellText: UnicodeString);
var
  Data: PShellObjectData;
begin
  with Sender as TVirtualStringTree do begin
    Data := Sender.GetNodeData(Node);
    case Column of
      -1: CellText := Data.Display;
    end;
  end;
end;

procedure TVirtualShellTree.VSTInitChildren(
  Sender: TBaseVirtualTree; Node: PVirtualNode; var ChildCount: Cardinal);
// Called just before a node with children (only folder nodes can have children) is expanded.
var
  Data,
  ChildData: PShellObjectData;
  ChildNode: PVirtualNode;
  ItemListPointer: PItemIDList;
  ChildPIDL: PItemIDList;
  SubFolder, Folder: IShellFolder;
  ShellInfo : TSHFileInfo;
  Enum: IEnumIDList;
  Dummy: ULONG;
  Attributes: UINT;

  // ILGetSize and ILCombine are included in version 5.0 or later of shell32.dll.
  // We re-implement them here, so that the VirtualShellTreeView works with
  // older verions of shell32.dll, too.

  function ILGetSize(Pidl: PItemIDList): ULONG;
  begin
    Result := 0;
    if Assigned(Pidl) then
      repeat
        Inc(Result, Pidl.mkid.cb);
        Inc(PByte(Pidl), Pidl.mkid.cb);
      until Pidl.mkid.cb = 0;
  end;

  function ILCombine(PidlAbsolut, PidlRelative: PItemIDList): PItemIDList;
  var
    SizeAbs,
    SizeRel,
    Sum: ULONG;
    P: PByte;
  begin
    Result := nil;
    SizeAbs := ILGetSize(PidlAbsolut);
    SizeRel := ILGetSize(PidlRelative);
    Sum := SizeAbs + SizeRel;
    if Sum > 0 then begin
      Result := Malloc.Alloc(Sum + 2);
      FillMemory(Result, Sum + 2, 0);
      P := Pointer(Result);
      if SizeAbs > 0 then begin
        CopyMemory(P, PidlAbsolut, SizeAbs);
        Inc(P, SizeAbs);
      end;
      if SizeRel > 0 then
        CopyMemory(P, PidlRelative, SizeRel);
    end;
  end;

begin
  Screen.Cursor := crHourGlass;
  try
    Sender.ValidateNode(Node, False);
    Data := Sender.GetNodeData(Node);
    ItemListPointer := Data.PIDL;
    Folder := Data.ShellFolder;

    if Folder.EnumObjects(0, SHCONTF_FOLDERS or SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN, Enum) = S_OK then 
      while Enum.Next(1, ChildPIDL, Dummy) = S_OK do begin
        try
          OleCheck(Folder.GetAttributesOf(1, ChildPIDL, Attributes));

          ChildNode := Sender.AddChild(Node);
          ChildData := Sender.GetNodeData(ChildNode);
        except
          // Free the memory allocated by EnumList.Next.
          Malloc.Free(ChildPIDL);
          raise;
        end;
        ChildData.PIDL := ILCombine(ItemListPointer, ChildPIDL);

        if Folder.BindToObject(ChildPIDL, nil, IID_IShellFolder, SubFolder) = NOERROR
          then ChildData.ShellFolder := SubFolder
          else ChildData.ShellFolder := nil;

        // Get details about the node via SHGetFileInfo
        OleCheck(SHGetFileInfo(PChar(ChildData.PIDL), 0, ShellInfo,
                      SizeOf(ShellInfo), SHGFI_PIDL or SHGFI_DISPLAYNAME or SHGFI_ATTRIBUTES	or SHGFI_SYSICONINDEX or SHGFI_SMALLICON));
        ChildData.Display := ShellInfo.szDisplayName;
        ChildData.Attributes := ShellInfo.dwAttributes;
        ChildData.CloseIndex := ShellInfo.iIcon;
        OleCheck(SHGetFileInfo(PChar(ChildData.PIDL), 0, ShellInfo,
                      SizeOf(ShellInfo), SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_OPENICON));
        ChildData.OpenIndex := ShellInfo.iIcon;

      end;

    ChildCount := Sender.ChildCount[Node];

    if ChildCount > 0 then
      Sender.Sort(Node, 0, sdAscending, False);

  finally
    Screen.Cursor := crDefault;
  end;

end;

procedure TVirtualShellTree.VSTGetImageIndex(
  Sender: TBaseVirtualTree; Node: PVirtualNode; Kind: TVTImageKind;
  Column: TColumnIndex; var Ghosted: Boolean; var ImageIndex: Integer);
// Returns the proper node image which has been determine on initialization time.
// Also overlay images are used properly for shared folders.
var
  Data: PShellObjectData;
begin
  if Column = -1 then
  begin
    Data := Sender.GetNodeData(Node);
    case Kind of
      ikNormal,
      ikSelected:
        begin
          if Sender.Expanded[Node] then
            ImageIndex := Data.OpenIndex
          else
            ImageIndex := Data.CloseIndex;
        end;
      ikOverlay:
        if (Data.Attributes and SFGAO_SHARE) <> 0 then
          ImageIndex := 0
        else
          if (Data.Attributes and SFGAO_LINK) <> 0 then
            ImageIndex := 1;
    end;
  end;
end;

procedure TVirtualShellTree.VSTCompareNodes(
  Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode;
  Column: TColumnIndex; var Result: Integer);
var
  Data1,
  Data2: PShellObjectData;
  Path1,
  Path2: array[0..MAX_PATH] of Char;
begin
  Data1 := Sender.GetNodeData(Node1);
  Data2 := Sender.GetNodeData(Node2);

  // Folder are always before files. Check if *both* are folders or *both*
  // are non-folders, but not different.
  if ((Data1.Attributes xor Data2.Attributes) and SFGAO_FOLDER) <> 0 then
  begin
    // One of both is a folder the other is a file.
    if (Data1.Attributes and SFGAO_FOLDER) <> 0 then
      Result := -1
    else
      Result := 1;
  end
  else
    // Both are of same type (folder or file). Just compare captions.
    // Note that we use ANSI comparison, while the strings are Unicode.
    // Since this will implicitely convert the captions to ANSI for
    // comparation it might happen that the sort order is wrong for names
    // which contain text in a language other than the current OS language.
    // A full blown Unicode comparison is beyond the scope here.
    if SHGetPathFromIDList(Data1.PIDL, Path1) and SHGetPathFromIDList(Data2.PIDL, Path2) then
      Result := AnsiCompareText(Path1, Path2);
end;

procedure TVirtualShellTree.VSTDblClick(Sender: TObject);
var
  Data: PShellObjectData;
  Path: array[0..MAX_PATH] of Char;
begin
  if Sender is TBaseVirtualTree then
    with Sender as TBaseVirtualTree do
      if Assigned(FocusedNode) then begin
        Data := GetNodeData(FocusedNode);
        if ((Data.Attributes and SFGAO_FILESYSTEM) <> 0) and  // filesystem but ...
            ((Data.Attributes and SFGAO_FOLDER ) = 0)  then   // ... not folder
          if SHGetPathFromIDList(Data.PIDL, Path) then
            DoFileDblClick(Path, Data.Attributes);
      end;
end;

initialization
  SHGetMalloc(Malloc);
  SHGetDesktopFolder(DesktopFolder);
finalization
  Malloc := nil;
end.
